% Main script for the interaction phase.
% - Gets current state from breve,
% - Uses the RNNPB to decide the next state
% - Uses the PID to decide new velocities for each joint
% - Sends new velocities to breve
% (- Also updates PB vectors)
%
% NOTE: This function must be run after breve has been started, while
% the simulation "Tiny Dancer.tz" is running.  A trained RNNPB named "net"
% must also be present in the workspace.

% Create a java communication objext
communication = javaObject('Breve');

% Set start angles
goalElbowL = 0;
goalElbowR = 0;

goalShoulderL = [1.5 -0.8 -1];
goalShoulderR = [1.5 0.8 1];


% Create java PID objects for each joint angle

% Left and right elbow angles
pidLe = javaObject('PID', 5, 3, 3, 1);
pidRe = javaObject('PID', 5, 3, 3, 1);

% Right shoulder angles
pidR1 = javaObject('PID', 5, 3, 3, 1);
pidR2 = javaObject('PID', 5, 3, 3, 1);
pidR3 = javaObject('PID', 5, 3, 3, 1);

% Left shoulder angles
pidL1 = javaObject('PID', 5, 3, 3, 1);
pidL2 = javaObject('PID', 5, 3, 3, 1);
pidL3 = javaObject('PID', 5, 3, 3, 1);

% Initialize variables that hold PB errors
dPB1 = zeros(1, net.size{5}, net.size{4});
ePB = zeros(1,net.size{5});
ePB_1 = zeros(1,net.size{5});
ePB_2 = zeros(1,net.size{5});
dpb = zeros(1, net.size{5});
pbError = zeros(net.size{5}, net.size{2});

% Initialize PB vectors
pb = 0.5*ones(1, net.size{5});
pbused = pb;

% Initialize context nodes
contexNodes = 0.5 * ones(1, net.size{3});

% Set the light ball's start position
lightBall = [-3 4.5 0];

% Send message to breve with light ball's position
communication.read('localhost', 7776, sprintf('move-light_%f_%f_%f', lightBall(1), lightBall(2), lightBall(3)));

for i = 1:500,
    
    % Read current left elbow angle from breve
    elbowL = str2double(communication.read('localhost', 7776, 'get-left-elbow-angle'));
    % Send error to left elbow PID, and get new velocity
    v = pidLe.getNewVelocity(goalElbowL-elbowL, -1, 1);
    
    % Read current right elbow angle from breve
    elbowR = str2double(communication.read('localhost', 7776, 'get-right-elbow-angle'));
	% Send error to right elbow PID, and get new velocity
    v2 = pidRe.getNewVelocity(goalElbowR-elbowR, -1, 1);
    
    % Read current right shoulder angles from breve
    eval(communication.read('localhost', 7776, sprintf('get-right-arm-angles')));
    p = [X Y Z];
    shoulderXR = X; % Store X-angle for later use
    % Send error to each right shoulder angle PID, and get new velocities
    shoulderRV(1) = pidR1.getNewVelocity(goalShoulderR(1) - p(1), -1, 1);
    shoulderRV(2) = pidR2.getNewVelocity(goalShoulderR(2) - p(2), -1, 1);
    shoulderRV(3) = pidR3.getNewVelocity(goalShoulderR(3) - p(3), -1, 1);
    
    % Read current left shoulder angles from breve
    eval(communication.read('localhost', 7776, sprintf('get-left-arm-angles')));
    p = [X Y Z];
    shoulderXL = X;% Store X-angle for later use
    % Send error to each left shoulder angle PID, and get new velocities
    shoulderLV(1) = pidL1.getNewVelocity(goalShoulderL(1) - p(1), -1, 1);
    shoulderLV(2) = pidL2.getNewVelocity(goalShoulderL(2) - p(2), -1, 1);
    shoulderLV(3) = pidL3.getNewVelocity(goalShoulderL(3) - p(3), -1, 1);
    
    % Send new velocities to breve
    communication.read('localhost', 7776, sprintf('set-left-elbow-velocity_%f', v));
    communication.read('localhost', 7776, sprintf('set-right-elbow-velocity_%f', v2));
    communication.read('localhost', 7776, sprintf('set-left-arm-velocities_%f_%f_%f', shoulderLV(1), shoulderLV(2), shoulderLV(3)));
    communication.read('localhost', 7776, sprintf('set-right-arm-velocities_%f_%f_%f', shoulderRV(1), shoulderRV(2), shoulderRV(3)));

    % Tell breve to iterate the simulation
    communication.read('localhost', 7776, 'matlab-iterate');
    
    % Read sensors:
    eyeR = str2double(communication.read('localhost', 7776, sprintf('get-right-eye'))) / 5;
    eyeL = str2double(communication.read('localhost', 7776, sprintf('get-left-eye'))) / 5;
    
    % Pause for 0.1 seconds (this is so that breve runs smoothly)
    pause(0.1);
    
    currentPattern = [eyeL eyeR shoulderXL/1.5 elbowL shoulderXR/1.5 elbowR];
   
    % Activate - we now have a new set of desired goals
    net = rnnpb_activate(net, currentPattern, pbused);
    
    % Set new elbow and shoulder goals  (which will be used by the PID in
    % the next step)
    goalElbowL = net.L{3}(4);
    goalElbowR = -net.L{3}(6);
    goalShoulderL = [net.L{3}(3)*1.5 -0.8 -1];
    goalShoulderR = [net.L{3}(5)*1.5 0.8 1];
  
    
    % Find errors so that the PB vectors can be updated
    %
    % NOTE: net.L{3} is the third layer in the network: in other
    % words, the predicted state from the activation, which we now want to
    % be the target for the next step. The error is thus measured by
    % the difference between the target and current state.
    oError = net.L{3}.*(1 - net.L{3}).*(net.L{3} - currentPattern);
    hError = oError * net.w{1}' .* net.L{2} .* (1-net.L{2});

    % Find PB error
    for pbnr=1:net.size{5},
        pbError(pbnr,:) = hError .* net.w{2}(net.size{1}+pbnr,:) .* (1-net.w{2}(net.size{1}+pbnr,:));
    end;

    ePB = -sum((net.w{2}(net.size{1}+1:net.size{1}+net.size{5},:) .* pbError)');
    dpb = dpb + ePB;
     
    % Integrate using Simpson's rule
     pb = pb + 0.01*(dpb + 4*ePB_1 + ePB_2);
     ePB_2 = ePB_1;
     ePB_1 = dpb;
    
    % Sigmoid function - the PB vector used in the next step
    pbused = 1./(1 + exp(-pb))

end;